/**************************************************************/
/*                                                            */
/* Module  : smqesmp3.c                                       */
/*                                                            */
/* Description  : Sample Exit for use with the MQSeries Link  */
/*                SAP R/3. This exit is designed for use with */
/*                an outbound server. The exit returns the    */
/*                MQ message descripor as normal but removes  */
/*                the MQSAPH link for R/3 header from the body*/
/*                of the message, returning only the IDOC data*/
/*                exported from SAP.                          */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Include some 'standard' C header files                     */
/**************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/**************************************************************/
/* Include the header files for both MQSeries and the Link    */
/**************************************************************/
#include "cmqc.h"
#include "smqc.h"

/**************************************************************/
/* Define prototypes for the exit entry points. There are     */
/* four of these viz.                                         */
/*   Initialise                                               */ 
/*   Execute                                                  */    
/*   Return                                                   */    
/*   Terminate                                                */
/* The typedefs for the entry points are contained in smqc.h. */
/**************************************************************/
smqUserExitInitialise Initialise ;
smqUserExitExecute    Execute ; 
smqUserExitReturn     Return ;
smqUserExitTerminate  Terminate ;
      
/**************************************************************/
/* Declare data global to this module.                        */
/**************************************************************/
SMQEXIT_DIRECTION  Direction;               /* Data direction */
PMQCHAR            pUserBuffer;             /* Ptr to buffer  */

/**************************************************************/
/*                                                            */
/* Routine  : Initialise                                      */
/*                                                            */
/* Description  : Initialisation routine for the sample exit. */
/*                This is called after initialisation of the  */
/*                server and is responsible for returning the */
/*                addresses of the other entry points. It may */
/*                also perform any additional processing      */
/*                needed for subsequent calls to the exit.    */
/*                Note that this entry point must be the      */
/*                first or only one 'exported'. See the .def  */
/*                or .exp file.                               */  
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure.                                 */
/*               pExitInitialise - pointer to an SMQEXIT_INIT */
/*                 structure.                                 */     
/*                                                            */
/**************************************************************/
void MQENTRY Initialise(PSMQEXIT_SETUP pExitSetup,
                        PSMQEXIT_INIT  pExitInitialise)
{
  /************************************************************/
  /* Set a default good result code                           */
  /************************************************************/
  pExitInitialise -> Result = SMQRC_OK ;

  /************************************************************/
  /* First we check that the structures are things we can     */
  /* handle. We only recognise version we compiled with.  If  */
  /* the structures are not the correct version then set the  */
  /* result field to terminate. This causes the exit handler  */
  /* to bypass calling the Execute and Return functions but   */
  /* it will call the Terminate function. If we wish to cause */
  /* the server itself to end then we should return           */
  /* SMQRC_TERMINATE_SERVER (the exit handler should still    */
  /* call the Terminate function).                            */                                               
  /************************************************************/
  if ((pExitSetup -> Version != SMQEXIT_SETUP_VERSION) ||
      (pExitInitialise -> Version != SMQEXIT_INIT_VERSION))
  {
    pExitInitialise -> Result = SMQRC_TERMINATE_EXIT ;
  }
  else
  {
    /**********************************************************/
    /* At a minimum we must return the addresses of the other */
    /* entry points for the exit. The init structure contains */
    /* the fields for these pointers.                         */
    /**********************************************************/
    pExitInitialise -> ExecEntryPoint = (void*)Execute ;
    pExitInitialise -> RetnEntryPoint = (void*)Return ;
    pExitInitialise -> TermEntryPoint = (void*)Terminate ;

    /**********************************************************/
    /* Store the direction in some global storage. We need to */
    /* know this in the Execute function. The direction is    */
    /* passed to us in the SMQEXIT_INIT control block.        */
    /**********************************************************/
    Direction = pExitInitialise -> Direction;
  }

  /************************************************************/
  /* Put the Result field set in the INIT structure into the  */
  /* ExitOptions field of the SETUP structure for possible    */
  /* use by the termination routine.                          */
  /************************************************************/
  pExitSetup -> ExitOptions =
      pExitInitialise -> Result ;
   
  /************************************************************/
  /* Now return to the exit handler.                          */
  /************************************************************/
  return ;
} 
  
/**************************************************************/
/*                                                            */
/* Routine  : Terminate                                       */
/*                                                            */
/* Description  : Termination routine for sample exit. This   */
/*                is called at termination of the server. It  */
/*                is responsible for performing any cleanup   */
/*                of resources acquired during previous exit  */
/*                processing.                                 */
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure.                                 */
/*               ExitRequest - The reason for termination.    */
/*                                                            */
/**************************************************************/
void MQENTRY Terminate(PSMQEXIT_SETUP pExitSetup,
                       SMQEXIT_REQUEST ExitTerminate)
{
  /************************************************************/
  /* ExitTerminate gives us the state of the exit when we are */
  /* called for termination. We don't actually do anything    */
  /* with this information in this exit.                      */
  /************************************************************/
  switch (ExitTerminate)
  {
    case SMQEXIT_REQUEST_EXIT:
          /********************************************************/
          /* We're ending 'cos a previous invocation of an exit   */
          /* function decided it was a good idea.                 */
          /********************************************************/
      break ;
    case SMQEXIT_REQUEST_SERVER:
          /********************************************************/
          /* The server is ending so we must follow suit. This    */
          /* is the 'normal' state of affairs.                    */
          /********************************************************/
      break ;
    case SMQEXIT_REQUEST_ERROR:
          /********************************************************/
          /* The server has encountered an error and so is ending */
          /********************************************************/    
      break ;
    default:
      /********************************************************/
      /* We have reached an unknown state!!                   */
      /********************************************************/
      break ;  
  }  
  return ;
}

/**************************************************************/
/*                                                            */
/* Routine  : Execute                                         */
/*                                                            */
/* Description  : Execution routine for a sample exit. This   */
/*                is called ....                              */
/*                From the inbound server immediately after   */
/*                the MQGET from the input queue. The         */
/*                MessageData field of the input structure    */
/*                points to the message as retrieved from the */
/*                input queue. For the purposes of this exit. */
/*                we expect this data to consist of an        */
/*                MQSAPH followed by the IDOC 'stuff', if     */
/*                we don't think it's in this format we say   */
/*                we can't translate the data.                */   
/*                From the outbound server immediately before */
/*                the MQPUT to the destination queue. The     */
/*                data is presented as two separate entities  */
/*                viz. an MQSAPH and the IDOC 'stuff'.        */
/*                In this sample we discard the MQSAPH and    */
/*                present a single buffer containing only the */
/*                IDOC data on exit.                          */
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure.                                 */
/*               pExitInput - pointer to an SMQEXIT_INPUT     */
/*                 structure. The structure contains details  */
/*                 about the data to be processed and a ptr   */
/*                 to the data. This structure is input only  */
/*                 and should not be changed by this routine. */
/*               pExitOutput - pointer to an SMQEXIT_OUTPUT   */
/*                 structure. This is used to contain info.   */
/*                 about the converted data.                  */
/*                                                            */
/**************************************************************/
void MQENTRY Execute(PSMQEXIT_SETUP pExitSetup,
                     PSMQEXIT_INPUT pExitInput,
                     PSMQEXIT_OUTPUT pExitOutput)
{
  PMQSAPH  pTHeader ;
  PMQMD pMD;
  
  /************************************************************/
  /* The required behaviour is dependent on the direction of  */
  /* the exit. On the inbound server we are given a single    */
  /* buffer which contains the message as retrieved from MQ.  */
  /* If we can we separate this data (logically) into Header  */
  /* and IDOC. On the outbound server we are presented with   */
  /* separate (logical) header and IDOC. We may combine these,*/
  /* if we wish, into a single buffer and return this to the  */
  /* server.                                                  */
  /************************************************************/
  switch (Direction)
  {
    case SMQEXIT_INBOUND:
      /********************************************************/
      /* First have a sneak preview of the contents of the    */
      /* input buffer to see if it matches our expectations   */
      /* i.e. it has an MQSAPH at the front. If it doesn't    */
      /* do nothing further with it and return to the exit    */
      /* handler with a Result of CONVERT_FAIL and let        */
      /* the server do its own thing with the data. If the    */
      /* data looks O.K. at first glance then proceed.        */
      /********************************************************/
      pTHeader = (PMQSAPH)(pExitInput -> MessageData) ;
      if ((memcmp(pTHeader -> StrucId,
                  &MQSAPH_IDENTIFIER,
                  SMQ_ID_LEN) != 0)  ||
          (pTHeader -> Version != SMQ_VERSION1) ||
          (pExitInput -> MessageDataLen <=
               sizeof(MQSAPH)))
      {
        pExitOutput -> Result = SMQRC_CONVERT_FAIL ;
      }
      else
      {
        /******************************************************/
        /* Now attempt to obtain a buffer large enough to     */
        /* hold the data passed to us. If we can't get the    */
        /* storage then we go no further.                     */
        /******************************************************/
        if ((pUserBuffer = malloc(pExitInput -> MessageDataLen))
                                 == NULL)
        {
          pExitOutput -> Result = SMQRC_CONVERT_FAIL ;
        }
        else
        {
          /****************************************************/
          /* Now copy the MQSAPH structure from the input     */
          /* buffer to the front of our own buffer and set    */
          /* the Header field of the output structure to      */
          /* point to it.                                     */
          /****************************************************/        
          memcpy(pUserBuffer,
                 pExitInput -> MessageData,
                 sizeof(MQSAPH));
          pExitOutput -> Header = (PMQSAPH)pUserBuffer;
          pTHeader = pExitOutput -> Header ;

          /****************************************************/
          /* Copy the 'IDOC' portion of the input data to our */
          /* buffer. Although the header and data are         */
          /* contiguous in the output buffer we present them  */
          /* on output as two logically separate buffers.     */
          /* Note that since we are separating the two the    */
          /* MessageDataLen field on output is the length of  */
          /* the IDOC portion only.                           */  
          /****************************************************/
          pExitOutput -> MessageData =
              pUserBuffer + sizeof(MQSAPH);
          pExitOutput -> MessageDataLen =
              pExitInput -> MessageDataLen - sizeof(MQSAPH) ;
          memcpy(pExitOutput -> MessageData,
                 (PMQCHAR)(pExitInput -> MessageData) +
                     sizeof(MQSAPH),
                 pExitOutput->MessageDataLen) ;
          /****************************************************/
          /* Now set the MessageDesc field on output to be    */
          /* the same as on input and say all is well.        */
          /****************************************************/ 
          pExitOutput -> MessageDesc =
              pExitInput -> MessageDesc ;
          pExitOutput -> Result = SMQRC_CONVERT_OK ;
        }
      };

          /********************************************************/
          /* We've now finished with the inbound data.            */
          /********************************************************/
      break;

    case SMQEXIT_OUTBOUND:
      /********************************************************/
      /* Try to acquire a buffer large enough to hold         */
      /* the IDOC. If we can't acquire                        */
      /* such then we return to the exit handler saying that  */
      /* we can't convert the data. Note that we assume the   */
      /* data is valid and do no checking of such.            */
      /********************************************************/ 
      if ((pUserBuffer = malloc(pExitInput -> MessageDataLen))
                           == NULL)
      {
        pExitOutput -> Result = SMQRC_CONVERT_FAIL ;
      }
      else
      {
        pExitOutput -> Header = NULL ;
        /******************************************************/
        /* Now copy the body of the data i.e. the IDOC        */
        /* 'stuff' to the output buffer. Set the length       */
        /* on output to be the length of the header + IDOC.   */
        /******************************************************/
        memcpy(pUserBuffer,
               pExitInput->MessageData,
               pExitInput->MessageDataLen) ;
        pExitOutput -> MessageData = pUserBuffer ;
        pExitOutput -> MessageDataLen =
            pExitInput -> MessageDataLen;
        /******************************************************/
        /* Now set the MessageDesc field on output to be the  */
        /* same as on input and alter the message format.     */
        /******************************************************/ 
        pExitOutput -> MessageDesc = pExitInput -> MessageDesc ;
        pExitOutput -> Result = SMQRC_CONVERT_OK ;
        pMD = pExitOutput -> MessageDesc;
        memcpy(pMD -> Format, "IDOC    ", MQ_FORMAT_LENGTH);
      };
      break;
    default:
      pExitOutput -> Result = SMQRC_TERMINATE_EXIT;
      break;
  }

  return ;
}

/**************************************************************/
/*                                                            */
/* Routine  : Return                                          */
/*                                                            */
/* Description  : Return routine for sample exit. This        */
/*                routine is called when the server has       */
/*                completed processing the data returned by   */
/*                the execute function. Its function is to    */
/*                cleanup, if necessary, any resources        */
/*                acquired by the execute function. For this  */
/*                sample exit we free any buffer acquired     */
/*                by the previous call to the execute         */
/*                routine.                                    */
/*                Note that this function is ony called when  */
/*                execute has returned CONVERT_OK.            */ 
/*                                                            */
/* Parameters  : pExitSetup - pointer to an SMQEXIT_SETUP     */
/*                 structure. This is the structure as        */
/*                 returned by the previous call to an exit   */
/*                 routine.                                   */
/*               pExitOutput - pointer to an SMQEXIT_OUTPUT   */
/*                 structure. This is the structure as        */
/*                 returned by the previous call to the       */
/*                                                            */
/**************************************************************/
void MQENTRY Return(PSMQEXIT_SETUP pExitSetup,
                    PSMQEXIT_OUTPUT pExitOutput)
{
  /************************************************************/
  /* Free the allocated buffer if any.                        */
  /************************************************************/
  if (pUserBuffer != NULL)
  {
    free (pUserBuffer) ;
  }

  /************************************************************/
  /* Set all variables in the output structure to be          */
  /* NULLs or zeroes.                                         */
  /************************************************************/
  pExitOutput -> Header = NULL ;
  pExitOutput -> MessageDesc = NULL ;
  pExitOutput -> MessageData = NULL ;
  pExitOutput -> MessageDataLen = 0 ;

  /************************************************************/
  /* Simply return to the exit handler saying all is o.k.     */
  /************************************************************/
  pExitOutput -> Result = SMQRC_OK ;
  return ;
}
